# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path as osp
import math
import pickle
import os


def build_yolo_transforms(params):
    from paddlex.det import transforms
    target_size = params.image_shape[0]
    use_mixup = params.use_mixup
    dt_list = []
    if use_mixup:
        dt_list.append(
            transforms.MixupImage(
                alpha=params.mixup_alpha,
                beta=params.mixup_beta,
                mixup_epoch=int(params.num_epochs * 25. / 27)))
    dt_list.extend([
        transforms.RandomDistort(
            brightness_range=params.brightness_range,
            brightness_prob=params.brightness_prob,
            contrast_range=params.contrast_range,
            contrast_prob=params.contrast_prob,
            saturation_range=params.saturation_range,
            saturation_prob=params.saturation_prob,
            hue_range=params.hue_range,
            hue_prob=params.hue_prob), transforms.RandomExpand(
                prob=params.expand_prob,
                fill_value=[float(int(x * 255)) for x in params.image_mean])
    ])
    crop_image = params.crop_image
    if crop_image:
        dt_list.append(transforms.RandomCrop())
    dt_list.extend([
        transforms.Resize(
            target_size=target_size, interp='RANDOM'),
        transforms.RandomHorizontalFlip(prob=params.horizontal_flip_prob),
        transforms.Normalize(
            mean=params.image_mean, std=params.image_std)
    ])
    train_transforms = transforms.Compose(dt_list)
    eval_transforms = transforms.Compose([
        transforms.Resize(
            target_size=target_size, interp='CUBIC'),
        transforms.Normalize(
            mean=params.image_mean, std=params.image_std),
    ])
    return train_transforms, eval_transforms


def build_rcnn_transforms(params):
    from paddlex.det import transforms
    short_size = min(params.image_shape)
    max_size = max(params.image_shape)
    train_transforms = transforms.Compose([
        transforms.RandomDistort(
            brightness_range=params.brightness_range,
            brightness_prob=params.brightness_prob,
            contrast_range=params.contrast_range,
            contrast_prob=params.contrast_prob,
            saturation_range=params.saturation_range,
            saturation_prob=params.saturation_prob,
            hue_range=params.hue_range,
            hue_prob=params.hue_prob),
        transforms.RandomHorizontalFlip(prob=params.horizontal_flip_prob),
        transforms.Normalize(
            mean=params.image_mean, std=params.image_std),
        transforms.ResizeByShort(
            short_size=short_size, max_size=max_size),
        transforms.Padding(coarsest_stride=32 if params.with_fpn else 1),
    ])
    eval_transforms = transforms.Compose([
        transforms.Normalize(), transforms.ResizeByShort(
            short_size=short_size, max_size=max_size),
        transforms.Padding(coarsest_stride=32 if params.with_fpn else 1)
    ])
    return train_transforms, eval_transforms


def build_voc_datasets(dataset_path, train_transforms, eval_transforms):
    import paddlex as pdx
    train_file_list = osp.join(dataset_path, 'train_list.txt')
    eval_file_list = osp.join(dataset_path, 'val_list.txt')
    label_list = osp.join(dataset_path, 'labels.txt')
    train_dataset = pdx.datasets.VOCDetection(
        data_dir=dataset_path,
        file_list=train_file_list,
        label_list=label_list,
        transforms=train_transforms,
        shuffle=True)
    eval_dataset = pdx.datasets.VOCDetection(
        data_dir=dataset_path,
        file_list=eval_file_list,
        label_list=label_list,
        transforms=eval_transforms)
    return train_dataset, eval_dataset


def build_coco_datasets(dataset_path, train_transforms, eval_transforms):
    import paddlex as pdx
    data_dir = osp.join(dataset_path, 'JPEGImages')
    train_ann_file = osp.join(dataset_path, 'train.json')
    eval_ann_file = osp.join(dataset_path, 'val.json')
    train_dataset = pdx.datasets.CocoDetection(
        data_dir=data_dir,
        ann_file=train_ann_file,
        transforms=train_transforms,
        shuffle=True)
    eval_dataset = pdx.datasets.CocoDetection(
        data_dir=data_dir, ann_file=eval_ann_file, transforms=eval_transforms)
    return train_dataset, eval_dataset


def build_optimizer(step_each_epoch, params):
    import paddle.fluid as fluid
    from paddle.fluid.regularizer import L2Decay
    learning_rate = params.learning_rate
    num_epochs = params.num_epochs
    lr_decay_epochs = params.lr_decay_epochs
    warmup_steps = params.warmup_steps
    warmup_start_lr = params.warmup_start_lr

    boundaries = [b * step_each_epoch for b in lr_decay_epochs]
    values = [
        learning_rate * (0.1**i) for i in range(len(lr_decay_epochs) + 1)
    ]
    lr = fluid.layers.piecewise_decay(boundaries=boundaries, values=values)
    lr = fluid.layers.linear_lr_warmup(
        learning_rate=lr,
        warmup_steps=warmup_steps,
        start_lr=warmup_start_lr,
        end_lr=learning_rate)
    factor = 1e-04 if params.model in ['FasterRCNN', 'MaskRCNN'] else 5e-04
    optimizer = fluid.optimizer.Momentum(
        learning_rate=lr, momentum=0.9, regularization=L2Decay(factor))
    return optimizer


def train(task_path, dataset_path, params):
    import paddlex as pdx
    pdx.log_level = 3
    if params.model in ['YOLOv3', 'PPYOLO']:
        train_transforms, eval_transforms = build_yolo_transforms(params)
    elif params.model in ['FasterRCNN', 'MaskRCNN']:
        train_transforms, eval_transforms = build_rcnn_transforms(params)
    if osp.exists(osp.join(dataset_path, 'JPEGImages')) and \
        osp.exists(osp.join(dataset_path, 'train.json')) and \
        osp.exists(osp.join(dataset_path, 'val.json')):
        train_dataset, eval_dataset = build_coco_datasets(
            dataset_path=dataset_path,
            train_transforms=train_transforms,
            eval_transforms=eval_transforms)
    elif osp.exists(osp.join(dataset_path, 'train_list.txt')) and \
        osp.exists(osp.join(dataset_path, 'val_list.txt')) and \
        osp.exists(osp.join(dataset_path, 'labels.txt')):
        train_dataset, eval_dataset = build_voc_datasets(
            dataset_path=dataset_path,
            train_transforms=train_transforms,
            eval_transforms=eval_transforms)
    step_each_epoch = train_dataset.num_samples // params.batch_size
    train_batch_size = params.batch_size
    save_interval_epochs = params.save_interval_epochs
    save_dir = osp.join(task_path, 'output')
    pretrain_weights = params.pretrain_weights

    optimizer = build_optimizer(step_each_epoch, params)
    detector = getattr(pdx.cv.models, params.model)
    num_classes = len(train_dataset.labels) if params.model in ['YOLOv3', 'PPYOLO'] else \
        len(train_dataset.labels) + 1
    sensitivities_path = params.sensitivities_path
    eval_metric_loss = params.eval_metric_loss
    if eval_metric_loss is None:
        eval_metric_loss = 0.05
    model = detector(num_classes=num_classes, backbone=params.backbone)
    if params.model in ['YOLOv3', 'PPYOLO']:
        model.train_random_shapes = params.random_shape_sizes
    if params.model == 'YOLOv3':
        model.train(
            num_epochs=params.num_epochs,
            train_dataset=train_dataset,
            train_batch_size=train_batch_size,
            eval_dataset=eval_dataset,
            save_interval_epochs=save_interval_epochs,
            log_interval_steps=2,
            save_dir=save_dir,
            pretrain_weights=pretrain_weights,
            optimizer=optimizer,
            use_vdl=True,
            sensitivities_file=sensitivities_path,
            eval_metric_loss=eval_metric_loss,
            resume_checkpoint=params.resume_checkpoint)
    else:
        model.train(
            num_epochs=params.num_epochs,
            train_dataset=train_dataset,
            train_batch_size=train_batch_size,
            eval_dataset=eval_dataset,
            save_interval_epochs=save_interval_epochs,
            log_interval_steps=2,
            save_dir=save_dir,
            pretrain_weights=pretrain_weights,
            optimizer=optimizer,
            use_vdl=True,
            resume_checkpoint=params.resume_checkpoint)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import os.path as osp
import shutil
from enum import Enum
import traceback
import chardet
from PIL import Image
import numpy as np
import json
from ..utils import set_folder_status, get_folder_status, DatasetStatus


def copy_directory(src, dst, files=None):
    """从src目录copy文件至dst目录，
       注意:拷贝前会先清空dst中的所有文件

    Args:
        src: 源目录路径
        dst: 目标目录路径
        files: 需要拷贝的文件列表（src的相对路径）
    """
    set_folder_status(dst, DatasetStatus.XCOPYING, os.getpid())
    if files is None:
        files = list_files(src)
    try:
        message = '{} {}'.format(os.getpid(), len(files))
        set_folder_status(dst, DatasetStatus.XCOPYING, message)
        if not osp.samefile(src, dst):
            for i, f in enumerate(files):
                items = osp.split(f)
                if len(items) > 2:
                    continue
                if len(items) == 2:
                    if not osp.isdir(osp.join(dst, items[0])):
                        if osp.exists(osp.join(dst, items[0])):
                            os.remove(osp.join(dst, items[0]))
                        os.makedirs(osp.join(dst, items[0]))
                shutil.copy(osp.join(src, f), osp.join(dst, f))
        set_folder_status(dst, DatasetStatus.XCOPYDONE)
    except Exception as e:
        error_info = traceback.format_exc()
        set_folder_status(dst, DatasetStatus.XCOPYFAIL, error_info)


def is_pic(filename):
    """ 判断文件是否为图片格式

    Args:
        filename: 文件路径
    """
    suffixes = {'JPEG', 'jpeg', 'JPG', 'jpg', 'BMP', 'bmp', 'PNG', 'png'}
    suffix = filename.strip().split('.')[-1]
    if suffix not in suffixes:
        return False
    return True


def replace_ext(filename, new_ext):
    """ 替换文件后缀

    Args:
        filename: 文件路径
        new_ext: 需要替换的新的后缀
    """
    items = filename.split(".")
    items[-1] = new_ext
    new_filename = ".".join(items)
    return new_filename


def get_encoding(filename):
    """ 获取文件编码方式

    Args:
        filename: 文件路径
    """
    f = open(filename, 'rb')
    data = f.read()
    file_encoding = chardet.detect(data).get('encoding')
    return file_encoding


def pil_imread(file_path):
    """ 获取分割标注图片信息

    Args:
        filename: 文件路径
    """
    im = Image.open(file_path)
    return np.asarray(im)


def check_list_txt(list_txts):
    """ 检查切分信息文件的格式

    Args:
        list_txts: 包含切分信息文件路径的list
    """
    for list_txt in list_txts:
        if not osp.exists(list_txt):
            continue
        with open(list_txt) as f:
            for line in f:
                items = line.strip().split()
                if len(items) != 2:
                    raise Exception('{} 格式错误. 列表应包含两列，由空格分离。'.format(list_txt))


def read_seg_ann(pngfile):
    """ 解析语义分割的标注png图片

    Args:
        pngfile: 包含标注信息的png图片路径
    """
    grt = pil_imread(pngfile)
    labels = list(np.unique(grt))
    if 255 in labels:
        labels.remove(255)
    return labels, grt.shape


def read_coco_ann(img_id, coco, cid2cname, catid2clsid):
    img_anno = coco.loadImgs(img_id)[0]
    im_w = float(img_anno['width'])
    im_h = float(img_anno['height'])

    ins_anno_ids = coco.getAnnIds(imgIds=img_id, iscrowd=0)
    instances = coco.loadAnns(ins_anno_ids)

    bboxes = []
    for inst in instances:
        x, y, box_w, box_h = inst['bbox']
        x1 = max(0, x)
        y1 = max(0, y)
        x2 = min(im_w - 1, x1 + max(0, box_w - 1))
        y2 = min(im_h - 1, y1 + max(0, box_h - 1))
        if inst['area'] > 0 and x2 >= x1 and y2 >= y1:
            inst['clean_bbox'] = [x1, y1, x2, y2]
            bboxes.append(inst)
        else:
            raise Exception("标注文件存在错误")
    num_bbox = len(bboxes)

    gt_bbox = np.zeros((num_bbox, 4), dtype=np.float32)
    gt_class = [""] * num_bbox
    gt_poly = [None] * num_bbox

    for i, box in enumerate(bboxes):
        catid = box['category_id']
        gt_class[i] = cid2cname[catid2clsid[catid]]
        gt_bbox[i, :] = box['clean_bbox']
        # is_crowd[i][0] = box['iscrowd']
        if 'segmentation' in box:
            gt_poly[i] = box['segmentation']

    anno_dict = {
        'h': im_h,
        'w': im_w,
        'gt_class': gt_class,
        'gt_bbox': gt_bbox,
        'gt_poly': gt_poly,
    }
    return anno_dict


def get_npy_from_coco_json(coco, npy_path, files):
    """ 从实例分割标注的json文件中，获取每张图片的信息，并存为npy格式

    Args:
        coco: 从json文件中解析出的标注信息
        npy_path: npy文件保存的地址
        files: 需要生成npy文件的目录
    """
    img_ids = coco.getImgIds()
    cat_ids = coco.getCatIds()
    anno_ids = coco.getAnnIds()
    catid2clsid = dict({catid: i for i, catid in enumerate(cat_ids)})
    cid2cname = dict({
        clsid: coco.loadCats(catid)[0]['name']
        for catid, clsid in catid2clsid.items()
    })
    iname2id = dict()
    for img_id in img_ids:
        img_name = osp.split(coco.loadImgs(img_id)[0]["file_name"])[-1]
        iname2id[img_name] = img_id

    if not osp.exists(npy_path):
        os.makedirs(npy_path)

    for img in files:
        img_id = iname2id[osp.split(img)[-1]]
        anno_dict = read_coco_ann(img_id, coco, cid2cname, catid2clsid)

        img_name = osp.split(img)[-1]
        npy_name = replace_ext(img_name, "npy")
        np.save(osp.join(npy_path, npy_name), anno_dict)


def get_label_count(label_info):
    """ 根据存储的label_info字段，计算label_count字段

    Args:
        label_info: 存储的label_info
    """
    label_count = dict()
    for key in sorted(label_info):
        label_count[key] = len(label_info[key])
    return label_count


class MyEncoder(json.JSONEncoder):
    # 调整json文件存储形式
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(MyEncoder, self).default(obj)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from paddle.fluid.contrib.slim.quantization.quantization_pass import QuantizationTransformPass
from paddle.fluid.contrib.slim.quantization.quantization_pass import AddQuantDequantPass
from paddle.fluid.contrib.slim.quantization.quantization_pass import _out_scale_op_list
from paddle.fluid.contrib.slim.quantization import PostTrainingQuantization
import paddlex.utils.logging as logging
import paddle.fluid as fluid
import os
import re
import numpy as np
import time


def _load_variable_data(scope, var_name):
    '''
    Load variable value from scope
    '''
    var_node = scope.find_var(var_name)
    assert var_node is not None, \
        "Cannot find " + var_name + " in scope."
    return np.array(var_node.get_tensor())


class PaddleXPostTrainingQuantization(PostTrainingQuantization):
    def __init__(self,
                 executor,
                 dataset,
                 program,
                 inputs,
                 outputs,
                 batch_size=10,
                 batch_nums=None,
                 scope=None,
                 algo="KL",
                 quantizable_op_type=["conv2d", "depthwise_conv2d", "mul"],
                 is_full_quantize=False,
                 is_use_cache_file=False,
                 cache_dir="./temp_post_training"):
        '''
        The class utilizes post training quantization methon to quantize the
        fp32 model. It uses calibrate data to calculate the scale factor of
        quantized variables, and inserts fake quant/dequant op to obtain the
        quantized model.
        Args:
            executor(fluid.Executor): The executor to load, run and save the
                quantized model.
            dataset(Python Iterator): The data Reader.
            program(fluid.Program): The paddle program, save the parameters for model.
            inputs(dict): The input of prigram.
            outputs(dict): The output of program.
            batch_size(int, optional): The batch size of DataLoader. Default is 10.
            batch_nums(int, optional): If batch_nums is not None, the number of
                calibrate data is batch_size*batch_nums. If batch_nums is None, use
                all data provided by sample_generator as calibrate data.
            scope(fluid.Scope, optional): The scope of the program, use it to load
                and save variables. If scope=None, get scope by global_scope().
            algo(str, optional): If algo=KL, use KL-divergenc method to
                get the more precise scale factor. If algo='direct', use
                abs_max methon to get the scale factor. Default is KL.
            quantizable_op_type(list[str], optional): List the type of ops
                that will be quantized. Default is ["conv2d", "depthwise_conv2d",
                "mul"].
            is_full_quantized(bool, optional): If set is_full_quantized as True,
                apply quantization to all supported quantizable op type. If set
                is_full_quantized as False, only apply quantization to the op type
                according to the input quantizable_op_type.
            is_use_cache_file(bool, optional): If set is_use_cache_file as False,
                all temp data will be saved in memory. If set is_use_cache_file as True,
                it will save temp data to disk. When the fp32 model is complex or
                the number of calibrate data is large, we should set is_use_cache_file
                as True. Defalut is False.
            cache_dir(str, optional): When is_use_cache_file is True, set cache_dir as
                the directory for saving temp data. Default is ./temp_post_training.
        Returns:
            None
        '''
        self._support_activation_quantize_type = [
            'range_abs_max', 'moving_average_abs_max', 'abs_max'
        ]
        self._support_weight_quantize_type = [
            'abs_max', 'channel_wise_abs_max'
        ]
        self._support_algo_type = ['KL', 'abs_max', 'min_max']
        self._support_quantize_op_type = \
            list(set(QuantizationTransformPass._supported_quantizable_op_type +
                AddQuantDequantPass._supported_quantizable_op_type))

        # Check inputs
        assert executor is not None, "The executor cannot be None."
        assert batch_size > 0, "The batch_size should be greater than 0."
        assert algo in self._support_algo_type, \
            "The algo should be KL, abs_max or min_max."

        self._executor = executor
        self._dataset = dataset
        self._batch_size = batch_size
        self._batch_nums = batch_nums
        self._scope = fluid.global_scope() if scope == None else scope
        self._algo = algo
        self._is_use_cache_file = is_use_cache_file
        self._cache_dir = cache_dir
        self._activation_bits = 8
        self._weight_bits = 8
        self._activation_quantize_type = 'range_abs_max'
        self._weight_quantize_type = 'channel_wise_abs_max'
        if self._is_use_cache_file and not os.path.exists(self._cache_dir):
            os.mkdir(self._cache_dir)

        if is_full_quantize:
            self._quantizable_op_type = self._support_quantize_op_type
        else:
            self._quantizable_op_type = quantizable_op_type
            for op_type in self._quantizable_op_type:
                assert op_type in self._support_quantize_op_type + \
                    AddQuantDequantPass._activation_type, \
                    op_type + " is not supported for quantization."

        self._place = self._executor.place
        self._program = program
        self._feed_list = list(inputs.values())
        self._fetch_list = list(outputs.values())
        self._data_loader = None

        self._out_scale_op_list = _out_scale_op_list
        self._bit_length = 8
        self._quantized_weight_var_name = set()
        self._quantized_act_var_name = set()
        self._sampling_data = {}
        self._quantized_var_kl_threshold = {}
        self._quantized_var_min = {}
        self._quantized_var_max = {}
        self._quantized_var_abs_max = {}

    def quantize(self):
        '''
        Quantize the fp32 model. Use calibrate data to calculate the scale factor of
        quantized variables, and inserts fake quant/dequant op to obtain the
        quantized model.
        Args:
            None
        Returns:
            the program of quantized model.
        '''
        self._load_model_data()
        self._collect_target_varnames()
        self._set_activation_persistable()
        batch_ct = 0
        for data in self._data_loader():
            batch_ct += 1
            if self._batch_nums and batch_ct >= self._batch_nums:
                break
        batch_id = 0
        logging.info("Start to run batch!")
        for data in self._data_loader():
            start = time.time()
            with fluid.scope_guard(self._scope):
                self._executor.run(program=self._program,
                                   feed=data,
                                   fetch_list=self._fetch_list,
                                   return_numpy=False)
            if self._algo == "KL":
                self._sample_data(batch_id)
            else:
                self._sample_threshold()
            end = time.time()
            logging.debug(
                '[Run batch data] Batch={}/{}, time_each_batch={} s.'.format(
                    str(batch_id + 1), str(batch_ct), str(end - start)))
            batch_id += 1
            if self._batch_nums and batch_id >= self._batch_nums:
                break
        logging.info("All run batch: ".format(batch_id))
        self._reset_activation_persistable()
        logging.info("Calculate scale factor ...")
        if self._algo == "KL":
            self._calculate_kl_threshold()
        logging.info("Update the program ...")
        if self._algo in ["KL", "abs_max"]:
            self._update_program()
        else:
            self._save_input_threhold()
        logging.info("Save ...")
        self._save_output_threshold()
        logging.info("Finish quant!")
        return self._program

    def save_quantized_model(self, save_model_path):
        '''
        Save the quantized model to the disk.
        Args:
            save_model_path(str): The path to save the quantized model
        Returns:
            None
        '''
        with fluid.scope_guard(self._scope):
            feed_vars_names = [var.name for var in self._feed_list]
            fluid.io.save_inference_model(
                dirname=save_model_path,
                feeded_var_names=feed_vars_names,
                target_vars=self._fetch_list,
                executor=self._executor,
                params_filename='__params__',
                main_program=self._program)

    def _load_model_data(self):
        '''
        Set data loader.
        '''
        feed_vars = [fluid.framework._get_var(var.name, self._program) \
            for var in self._feed_list]
        self._data_loader = fluid.io.DataLoader.from_generator(
            feed_list=feed_vars, capacity=3 * self._batch_size, iterable=True)
        self._data_loader.set_sample_list_generator(
            self._dataset.generator(
                self._batch_size, drop_last=True),
            places=self._place)

    def _calculate_kl_threshold(self):
        '''
        Calculate the KL threshold of quantized variables.
        '''
        assert self._algo == "KL", "The algo should be KL to calculate kl threshold."
        ct = 1
        # Abs_max threshold for weights
        for var_name in self._quantized_weight_var_name:
            start = time.time()
            weight_data = self._sampling_data[var_name]
            weight_threshold = None
            if self._weight_quantize_type == "abs_max":
                weight_threshold = np.max(np.abs(weight_data))
            elif self._weight_quantize_type == "channel_wise_abs_max":
                weight_threshold = []
                for i in range(weight_data.shape[0]):
                    abs_max_value = np.max(np.abs(weight_data[i]))
                    weight_threshold.append(abs_max_value)
            self._quantized_var_kl_threshold[var_name] = weight_threshold
            end = time.time()
            logging.debug(
                '[Calculate weight] Weight_id={}/{}, time_each_weight={} s.'.
                format(
                    str(ct),
                    str(len(self._quantized_weight_var_name)),
                    str(end - start)))
            ct += 1

        ct = 1
        # KL threshold for activations
        if self._is_use_cache_file:
            for var_name in self._quantized_act_var_name:
                start = time.time()
                sampling_data = []
                file_name = os.path.join(self._cache_dir, var_name)
                cache_dir, var_name_ = os.path.split(file_name)
                filenames = [f for f in os.listdir(cache_dir) \
                    if re.match(var_name_ + '_[0-9]+.npy', f)]
                for filename in filenames:
                    file_path = os.path.join(cache_dir, filename)
                    sampling_data.append(np.load(file_path))
                    os.remove(file_path)
                sampling_data = np.concatenate(sampling_data)
                self._quantized_var_kl_threshold[var_name] = \
                    self._get_kl_scaling_factor(np.abs(sampling_data))
                end = time.time()
                logging.debug(
                    '[Calculate activation] Activation_id={}/{}, time_each_activation={} s.'.
                    format(
                        str(ct),
                        str(len(self._quantized_act_var_name)),
                        str(end - start)))
                ct += 1
        else:
            for var_name in self._quantized_act_var_name:
                start = time.time()
                self._sampling_data[var_name] = np.concatenate(
                    self._sampling_data[var_name])
                self._quantized_var_kl_threshold[var_name] = \
                    self._get_kl_scaling_factor(np.abs(self._sampling_data[var_name]))
                end = time.time()
                logging.debug(
                    '[Calculate activation] Activation_id={}/{}, time_each_activation={} s.'.
                    format(
                        str(ct),
                        str(len(self._quantized_act_var_name)),
                        str(end - start)))
                ct += 1

    def _sample_data(self, iter):
        '''
        Sample the tensor data of quantized variables,
        applied in every iteration.
        '''
        assert self._algo == "KL", "The algo should be KL to sample data."
        for var_name in self._quantized_weight_var_name:
            if var_name not in self._sampling_data:
                var_tensor = _load_variable_data(self._scope, var_name)
                self._sampling_data[var_name] = var_tensor

        if self._is_use_cache_file:
            for var_name in self._quantized_act_var_name:
                var_tensor = _load_variable_data(self._scope, var_name)
                var_tensor = var_tensor.ravel()
                save_path = os.path.join(self._cache_dir,
                                         var_name + "_" + str(iter) + ".npy")
                save_dir, file_name = os.path.split(save_path)
                if not os.path.exists(save_dir):
                    os.makedirs(save_dir)
                np.save(save_path, var_tensor)
        else:
            for var_name in self._quantized_act_var_name:
                if var_name not in self._sampling_data:
                    self._sampling_data[var_name] = []
                var_tensor = _load_variable_data(self._scope, var_name)
                var_tensor = var_tensor.ravel()
                self._sampling_data[var_name].append(var_tensor)


class PaddleXPostTrainingQuantizationV2(PostTrainingQuantization):
    def __init__(self,
                 executor,
                 dataset,
                 program,
                 inputs,
                 outputs,
                 batch_size=10,
                 batch_nums=None,
                 scope=None,
                 algo="KL",
                 quantizable_op_type=["conv2d", "depthwise_conv2d", "mul"],
                 is_full_quantize=False,
                 activation_bits=8,
                 weight_bits=8,
                 activation_quantize_type='range_abs_max',
                 weight_quantize_type='channel_wise_abs_max',
                 optimize_model=False,
                 is_use_cache_file=False,
                 cache_dir="./temp_post_training"):
        '''
        Constructor.

        Args:
            executor(fluid.Executor): The executor to load, run and save the
                quantized model.
            dataset(Python Iterator): The data Reader.
            program(fluid.Program): The paddle program, save the parameters for model.
            inputs(dict): The input of prigram.
            outputs(dict): The output of program.

            scope(fluid.Scope, optional): The scope of the program, use it to load
                and save variables. If scope=None, get scope by global_scope().
            batch_size(int, optional): The batch size of DataLoader. Default is 10.
            batch_nums(int, optional): If batch_nums is not None, the number of
                calibrate data is batch_size*batch_nums. If batch_nums is None, use
                all data provided by sample_generator as calibrate data.
            algo(str, optional): If algo='KL', use KL-divergenc method to
                get the KL threshold for quantized activations and get the abs_max
                value for quantized weights. If algo='abs_max', get the abs max
                value for activations and weights. If algo= 'min_max', get the min
                and max value for quantized activations and weights. Default is KL.
            quantizable_op_type(list[str], optional): List the type of ops
                that will be quantized. Default is ["conv2d", "depthwise_conv2d",
                "mul"].
            is_full_quantized(bool, optional): If set is_full_quantized as True,
                apply quantization to all supported quantizable op type. If set
                is_full_quantized as False, only apply quantization to the op type
                according to the input quantizable_op_type.
            activation_bits(int): quantization bit number for activation.
            weight_bits(int, optional): quantization bit number for weights.
            activation_quantize_type(str): quantization type for activation,
                now support 'range_abs_max', 'moving_average_abs_max' and 'abs_max'.
                This param only specifies the fake ops in saving quantized model.
                If it is 'range_abs_max' or 'moving_average_abs_max', we save the scale
                obtained by post training quantization in fake ops. Note that, if it
                is 'abs_max', the scale will not be saved in fake ops.
            weight_quantize_type(str): quantization type for weights,
                support 'abs_max' and 'channel_wise_abs_max'. This param only specifies
                the fake ops in saving quantized model, and we save the scale obtained
                by post training quantization in fake ops. Compared to 'abs_max',
                the model accuracy is usually higher when it is 'channel_wise_abs_max'.
            optimize_model(bool, optional): If set optimize_model as True, it applies
                some passes to the model before quantization, and it supports
                `conv2d/depthwise_conv2d + bn` pass so far. Some targets require the
                weights are quantized by tensor-wise method, which means the weights
                scale for all channel are the same. However, if fuse
                `conv2d/depthwise_conv2d + bn`, the weights scale for all channel will
                be different. In address this problem, fuse the pattern before
                quantization. Default False.
            is_use_cache_file(bool, optional): This param is deprecated.
            cache_dir(str, optional): This param is deprecated.
        Returns:
            None

        '''

        self._support_activation_quantize_type = [
            'range_abs_max', 'moving_average_abs_max', 'abs_max'
        ]
        self._support_weight_quantize_type = [
            'abs_max', 'channel_wise_abs_max'
        ]
        self._support_algo_type = ['KL', 'abs_max', 'min_max']
        self._dynamic_quantize_op_type = ['lstm']
        self._support_quantize_op_type = \
            list(set(QuantizationTransformPass._supported_quantizable_op_type +
                AddQuantDequantPass._supported_quantizable_op_type +
                self._dynamic_quantize_op_type))

        # Check inputs
        assert executor is not None, "The executor cannot be None."

        assert batch_size > 0, "The batch_size should be greater than 0."
        assert algo in self._support_algo_type, \
            "The algo should be KL, abs_max or min_max."
        assert activation_quantize_type in self._support_activation_quantize_type, \
            "The activation_quantize_type ({}) should in ({}).".format(
            activation_quantize_type, self._support_activation_quantize_type)
        assert weight_quantize_type in self._support_weight_quantize_type, \
            "The weight_quantize_type ({}) shoud in ({}).".format(
            weight_quantize_type, self._support_weight_quantize_type)

        # Save input params
        self._executor = executor
        self._dataset = dataset
        self._scope = fluid.global_scope() if scope == None else scope
        self._batch_size = batch_size
        self._batch_nums = batch_nums
        self._algo = algo
        self._activation_bits = activation_bits
        self._weight_bits = weight_bits
        self._activation_quantize_type = activation_quantize_type
        self._weight_quantize_type = weight_quantize_type
        self._is_full_quantize = is_full_quantize
        if is_full_quantize:
            self._quantizable_op_type = self._support_quantize_op_type
        else:
            self._quantizable_op_type = quantizable_op_type
            for op_type in self._quantizable_op_type:
                assert op_type in self._support_quantize_op_type, \
                    op_type + " is not supported for quantization."
        self._optimize_model = optimize_model

        # Define variables
        self._place = self._executor.place
        self._program = program
        self._feed_list = [var.name for var in inputs.values()]
        self._fetch_list = list(outputs.values())
        self._data_loader = None

        self._out_scale_op_list = _out_scale_op_list
        self._quantized_weight_var_name = set()
        self._quantized_act_var_name = set()
        self._weight_op_pairs = {}
        # The vars for alog = KL
        self._sampling_act_abs_min_max = {}
        self._sampling_act_histogram = {}
        self._sampling_data = {}
        self._quantized_var_kl_threshold = {}
        self._histogram_bins = 2048
        # The vars for algo = min_max
        self._quantized_var_min = {}
        self._quantized_var_max = {}
        # The vars for algo = abs_max
        self._quantized_var_abs_max = {}

    def _load_model_data(self):
        '''
        Set data loader.
        '''
        logging.info("Set data loader ...")
        if self._program.num_blocks > 1:
            _logger.error("The post training quantization requires that the "
                          "program only has one block.")

        if self._optimize_model:
            self._optimize_fp32_model()

        feed_vars = [fluid.framework._get_var(var_name, self._program) \
            for var_name in self._feed_list]
        self._data_loader = fluid.io.DataLoader.from_generator(
            feed_list=feed_vars, capacity=3 * self._batch_size, iterable=True)
        self._data_loader.set_sample_list_generator(
            self._dataset.generator(
                self._batch_size, drop_last=True),
            places=self._place)

    def save_quantized_model(self, save_model_path):
        '''
        Save the quantized model to the disk.

        Args:
            save_model_path(str): The path to save the quantized model.
        Returns:
            None
        '''
        with fluid.scope_guard(self._scope):
            fluid.io.save_inference_model(
                dirname=save_model_path,
                model_filename='__model__',
                params_filename='__params__',
                feeded_var_names=self._feed_list,
                target_vars=self._fetch_list,
                executor=self._executor,
                main_program=self._program)
        logging.info("The quantized model is saved in " + save_model_path)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import platform
import os


class Params(object):
    def __init__(self):
        self.init_train_params()
        self.init_transform_params()

    def init_train_params(self):
        self.cuda_visible_devices = ''
        self.batch_size = 2
        self.save_interval_epochs = 1
        self.pretrain_weights = 'IMAGENET'
        self.model = 'MobileNetV2'
        self.num_epochs = 4
        self.learning_rate = 0.000125
        self.lr_decay_epochs = [2, 3]
        self.train_num = 0
        self.resume_checkpoint = None
        self.sensitivities_path = None
        self.eval_metric_loss = None

    def init_transform_params(self):
        self.image_shape = [224, 224]
        self.image_mean = [0.485, 0.456, 0.406]
        self.image_std = [0.229, 0.224, 0.225]
        self.horizontal_flip_prob = 0.5
        self.brightness_range = 0.9
        self.brightness_prob = 0.5
        self.contrast_range = 0.9
        self.contrast_prob = 0.5
        self.saturation_range = 0.9
        self.saturation_prob = 0.5
        self.hue_range = 18
        self.hue_prob = 0.5
        self.horizontal_flip = True
        self.brightness = True
        self.contrast = True
        self.saturation = True
        self.hue = True

    def load_from_dict(self, params_dict):
        for attr in params_dict:
            if hasattr(self, attr):
                method = getattr(self, "set_" + attr)
                method(params_dict[attr])

    def set_cuda_visible_devices(self, cuda_visible_devices):
        self.cuda_visible_devices = cuda_visible_devices

    def set_batch_size(self, batch_size):
        self.batch_size = batch_size

    def set_save_interval_epochs(self, save_interval_epochs):
        self.save_interval_epochs = save_interval_epochs

    def set_pretrain_weights(self, pretrain_weights):
        self.pretrain_weights = pretrain_weights

    def set_model(self, model):
        self.model = model

    def set_num_epochs(self, num_epochs):
        self.num_epochs = num_epochs

    def set_learning_rate(self, learning_rate):
        self.learning_rate = learning_rate

    def set_lr_decay_epochs(self, lr_decay_epochs):
        self.lr_decay_epochs = lr_decay_epochs

    def set_resume_checkpoint(self, resume_checkpoint):
        self.resume_checkpoint = resume_checkpoint

    def set_sensitivities_path(self, sensitivities_path):
        self.sensitivities_path = sensitivities_path

    def set_eval_metric_loss(self, eval_metric_loss):
        self.eval_metric_loss = eval_metric_loss

    def set_image_shape(self, image_shape):
        self.image_shape = image_shape

    def set_image_mean(self, image_mean):
        self.image_mean = image_mean

    def set_image_std(self, image_std):
        self.image_std = image_std

    def set_horizontal_flip(self, horizontal_flip):
        self.horizontal_flip = horizontal_flip
        if not horizontal_flip:
            self.horizontal_flip_prob = 0.0

    def set_horizontal_flip_prob(self, horizontal_flip_prob):
        if self.horizontal_flip:
            self.horizontal_flip_prob = horizontal_flip_prob

    def set_brightness_range(self, brightness_range):
        self.brightness_range = brightness_range

    def set_brightness_prob(self, brightness_prob):
        if self.brightness:
            self.brightness_prob = brightness_prob

    def set_brightness(self, brightness):
        self.brightness = brightness
        if not brightness:
            self.brightness_prob = 0.0

    def set_contrast(self, contrast):
        self.contrast = contrast
        if not contrast:
            self.contrast_prob = 0.0

    def set_contrast_prob(self, contrast_prob):
        if self.contrast:
            self.contrast_prob = contrast_prob

    def set_contrast_range(self, contrast_range):
        self.contrast_range = contrast_range

    def set_saturation(self, saturation):
        self.saturation = saturation
        if not saturation:
            self.saturation_prob = 0.0

    def set_saturation_prob(self, saturation_prob):
        if self.saturation:
            self.saturation_prob = saturation_prob

    def set_saturation_range(self, saturation_range):
        self.saturation_range = saturation_range

    def set_hue(self, hue):
        self.hue = hue
        if not hue:
            self.hue_prob = 0.0

    def set_hue_prob(self, hue_prob):
        if self.hue_prob:
            self.hue_prob = hue_prob

    def set_hue_range(self, hue_range):
        self.hue_range = hue_range

    def set_train_num(self, train_num):
        self.train_num = train_num


class ClsParams(Params):
    def __init__(self):
        super(ClsParams, self).__init__()
        self.lr_policy = 'Piecewise'
        self.vertical_flip_prob = 0.0
        self.vertical_flip = True
        self.rotate_prob = 0.0
        self.rotate_range = 30
        self.rotate = True

    def set_lr_policy(self, lr_policy):
        self.lr_policy = lr_policy

    def set_vertical_flip(self, vertical_flip):
        self.vertical_flip = vertical_flip
        if not self.vertical_flip:
            self.vertical_flip_prob = 0.0

    def set_vertical_flip_prob(self, vertical_flip_prob):
        if self.vertical_flip:
            self.vertical_flip_prob = vertical_flip_prob

    def set_rotate(self, rotate):
        self.rotate = rotate
        if not rotate:
            self.rotate_prob = 0.0

    def set_rotate_prob(self, rotate_prob):
        if self.rotate:
            self.rotate_prob = rotate_prob

    def set_rotate_range(self, rotate_range):
        self.rotate_range = rotate_range


class DetParams(Params):
    def __init__(self):
        super(DetParams, self).__init__()
        self.warmup_steps = 10
        self.warmup_start_lr = 0.
        self.use_mixup = True
        self.mixup_alpha = 1.5
        self.mixup_beta = 1.5
        self.expand_prob = 0.5
        self.expand_image = True
        self.crop_image = True
        self.backbone = 'ResNet18'
        self.model = 'FasterRCNN'
        self.with_fpn = True
        self.random_shape = True
        self.random_shape_sizes = [
            320, 352, 384, 416, 448, 480, 512, 544, 576, 608
        ]

    def set_warmup_steps(self, warmup_steps):
        self.warmup_steps = warmup_steps

    def set_warmup_start_lr(self, warmup_start_lr):
        self.warmup_start_lr = warmup_start_lr

    def set_use_mixup(self, use_mixup):
        self.use_mixup = use_mixup

    def set_mixup_alpha(self, mixup_alpha):
        self.mixup_alpha = mixup_alpha

    def set_mixup_beta(self, mixup_beta):
        self.mixup_beta = mixup_beta

    def set_expand_image(self, expand_image):
        self.expand_image = expand_image
        if not expand_image:
            self.expand_prob = 0.0

    def set_expand_prob(self, expand_prob):
        if self.expand_image:
            self.expand_prob = expand_prob

    def set_crop_image(self, crop_image):
        self.crop_image = crop_image

    def set_backbone(self, backbone):
        self.backbone = backbone

    def set_with_fpn(self, with_fpn):
        self.with_fpn = with_fpn

    def set_random_shape(self, random_shape):
        self.random_shape = random_shape

    def set_random_shape_sizes(self, random_shape_sizes):
        self.random_shape_sizes = random_shape_sizes


class SegParams(Params):
    def __init__(self):
        super(SegParams, self).__init__()
        self.loss_type = [True, True]
        self.lr_policy = 'Piecewise'
        self.optimizer = 'Adam'
        self.backbone = 'MobileNetV2_x1.0'
        self.blur = True
        self.blur_prob = 0.
        self.rotate = False
        self.max_rotation = 15
        self.scale_aspect = False
        self.min_ratio = 0.5
        self.aspect_ratio = 0.33
        self.vertical_flip_prob = 0.0
        self.vertical_flip = True
        self.model = 'UNet'

    def set_loss_type(self, loss_type):
        self.loss_type = loss_type

    def set_lr_policy(self, lr_policy):
        self.lr_policy = lr_policy

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer

    def set_backbone(self, backbone):
        self.backbone = backbone

    def set_blur(self, blur):
        self.blur = blur
        if not blur:
            self.blur_prob = 0.

    def set_blur_prob(self, blur_prob):
        if self.blur:
            self.blur_prob = blur_prob

    def set_rotate(self, rotate):
        self.rotate = rotate

    def set_max_rotation(self, max_rotation):
        self.max_rotation = max_rotation

    def set_scale_aspect(self, scale_aspect):
        self.scale_aspect = scale_aspect

    def set_min_ratio(self, min_ratio):
        self.min_ratio = min_ratio

    def set_aspect_ratio(self, aspect_ratio):
        self.aspect_ratio = aspect_ratio

    def set_vertical_flip(self, vertical_flip):
        self.vertical_flip = vertical_flip
        if not vertical_flip:
            self.vertical_flip_prob = 0.0

    def set_vertical_flip_prob(self, vertical_flip_prob):
        if vertical_flip_prob:
            self.vertical_flip_prob = vertical_flip_prob


PARAMS_CLASS_LIST = [ClsParams, DetParams, SegParams, DetParams, SegParams]


def recommend_parameters(params, train_nums, class_nums, memory_size_per_gpu):
    model_type = params['model']
    gpu_list = params['cuda_visible_devices']
    if 'cpu_num' in params:
        cpu_num = params['cpu_num']
    else:
        cpu_num = int(os.environ.get('CPU_NUM', 1))
        if cpu_num > 8:
            os.environ['CPU_NUM'] = '8'
    if not params['use_gpu']:
        gpu_nums = 0
    else:
        gpu_nums = len(gpu_list.split(','))

    # set batch_size
    if gpu_nums == 0 or platform.platform().startswith("Darwin"):
        if model_type.startswith('MobileNet'):
            batch_size = 8 * cpu_num
        elif model_type.startswith('DenseNet') or model_type.startswith('ResNet') \
            or model_type.startswith('Xception') or model_type.startswith('DarkNet') \
            or model_type.startswith('ShuffleNet'):
            batch_size = 4 * cpu_num
        elif model_type.startswith('YOLOv3') or model_type.startswith(
                'PPYOLO'):
            batch_size = 2 * cpu_num
        elif model_type.startswith('FasterRCNN') or model_type.startswith(
                'MaskRCNN'):
            batch_size = 1 * cpu_num
        elif model_type.startswith('DeepLab') or model_type.startswith('UNet') \
            or model_type.startswith('HRNet_W18') or model_type.startswith('FastSCNN'):
            batch_size = 2 * cpu_num
    else:
        if model_type.startswith('MobileNet'):
            batch_size = (memory_size_per_gpu - 513) // 57 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 125)
        elif model_type.startswith('DenseNet') or model_type.startswith('ResNet') \
            or model_type.startswith('Xception') or model_type.startswith('DarkNet') \
            or model_type.startswith('ShuffleNet'):
            batch_size = (memory_size_per_gpu - 739) // 211 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 16)
        elif model_type.startswith('YOLOv3'):
            batch_size = (memory_size_per_gpu - 1555) // 943 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 8)
        elif model_type.startswith('PPYOLO'):
            batch_size = (memory_size_per_gpu - 1691) // 1025 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 8)
        elif model_type.startswith('FasterRCNN'):
            batch_size = (memory_size_per_gpu - 1755) // 915 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 2)
        elif model_type.startswith('MaskRCNN'):
            batch_size = (memory_size_per_gpu - 2702) // 1188 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 2)
        elif model_type.startswith('DeepLab'):
            batch_size = (memory_size_per_gpu - 1469) // 1605 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 4)
        elif model_type.startswith('UNet'):
            batch_size = (memory_size_per_gpu - 1275) // 1256 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 4)
        elif model_type.startswith('HRNet_W18'):
            batch_size = (memory_size_per_gpu - 800) // 682 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 4)
        elif model_type.startswith('FastSCNN'):
            batch_size = (memory_size_per_gpu - 636) // 144 * gpu_nums
            batch_size = min(batch_size, gpu_nums * 4)

    if batch_size > train_nums // 2:
        batch_size = train_nums // 2
        gpu_list = '{}'.format(gpu_list.split(',')[0]) if gpu_nums > 0 else ''
    if batch_size <= 0:
        batch_size = 1

    # set learning_rate
    if model_type.startswith('MobileNet'):
        lr = (batch_size / 500.0) * 0.1
    elif model_type.startswith('DenseNet') or model_type.startswith('ResNet') \
        or model_type.startswith('Xception') or model_type.startswith('DarkNet') \
        or model_type.startswith('ShuffleNet'):
        lr = (batch_size / 256.0) * 0.1
    elif model_type.startswith('YOLOv3') or model_type.startswith('PPYOLO'):
        lr = 0.001 * batch_size / 64
        num_steps_each_epoch = train_nums // batch_size
        min_warmup_step = max(3 * num_steps_each_epoch, 50 * class_nums)
        if gpu_nums == 0:
            gpu_nums = 1
        warmup_step = min(min_warmup_step, int(400 * class_nums / gpu_nums))
    elif model_type.startswith('FasterRCNN') or model_type.startswith(
            'MaskRCNN'):
        lr = 0.02 * batch_size / 16
        num_steps_each_epoch = train_nums // batch_size
        min_warmup_step = max(num_steps_each_epoch, 50)
        if gpu_nums == 0:
            gpu_nums = 1
        warmup_step = min(min_warmup_step, int(4000 / gpu_nums))
    elif model_type.startswith('DeepLab') or model_type.startswith('UNet') \
        or model_type.startswith('HRNet_W18') or model_type.startswith('FastSCNN'):
        lr = 0.01 * batch_size / 2
        loss_type = [False, False]

    params['batch_size'] = batch_size
    params['learning_rate'] = lr
    params['cuda_visible_devices'] = gpu_list
    if model_type in ['YOLOv3', 'PPYOLO', 'FasterRCNN', 'MaskRCNN']:
        num_epochs = params['num_epochs']
        lr_decay_epochs = params['lr_decay_epochs']
        if warmup_step >= lr_decay_epochs[0] * num_steps_each_epoch:
            for i in range(len(lr_decay_epochs)):
                lr_decay_epochs[i] += warmup_step // num_steps_each_epoch
            num_epochs += warmup_step // num_steps_each_epoch
        params['num_epochs'] = num_epochs
        params['lr_decay_epochs'] = lr_decay_epochs
        params['warmup_steps'] = warmup_step
    if 'loss_type' in params:
        params['loss_type'] = loss_type



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def get_base_params(model_type, per_gpu_memory, num_train_samples, num_gpu,
                    gpu_list, num_classes):
    params = dict()
    params["model"] = model_type
    params["cpu_num"] = 1
    if num_gpu == 0:
        batch_size = 4
        params['cuda_visible_devices'] = ''
    else:
        params['cuda_visible_devices'] = str(gpu_list).strip("[]")
        if model_type.startswith('MobileNet'):
            batch_size = (per_gpu_memory - 513) // 57 * num_gpu
            batch_size = min(batch_size, num_gpu * 125)
        elif model_type.startswith('DenseNet') or model_type.startswith('ResNet') \
            or model_type.startswith('Xception') or model_type.startswith('DarkNet') \
            or model_type.startswith('ShuffleNet'):
            batch_size = (per_gpu_memory - 739) // 211 * num_gpu
            batch_size = min(batch_size, num_gpu * 16)
        elif model_type.startswith('YOLOv3'):
            batch_size = (per_gpu_memory - 1555) // 943 * num_gpu
            batch_size = min(batch_size, num_gpu * 8)
        elif model_type.startswith('PPYOLO'):
            batch_size = (per_gpu_memory - 1691) // 1025 * num_gpu
            batch_size = min(batch_size, num_gpu * 8)
        elif model_type.startswith('FasterRCNN'):
            batch_size = (per_gpu_memory - 1755) // 915 * num_gpu
            batch_size = min(batch_size, num_gpu * 2)
        elif model_type.startswith('MaskRCNN'):
            batch_size = (per_gpu_memory - 2702) // 1188 * num_gpu
            batch_size = min(batch_size, num_gpu * 2)
        elif model_type.startswith('DeepLab'):
            batch_size = (per_gpu_memory - 1469) // 1605 * num_gpu
            batch_size = min(batch_size, num_gpu * 4)
        elif model_type.startswith('UNet'):
            batch_size = (per_gpu_memory - 1275) // 1256 * num_gpu
            batch_size = min(batch_size, num_gpu * 4)
        elif model_type.startswith('HRNet_W18'):
            batch_size = (per_gpu_memory - 800) // 682 * num_gpu
            batch_size = min(batch_size, num_gpu * 4)
        elif model_type.startswith('FastSCNN'):
            batch_size = (per_gpu_memory - 636) // 144 * num_gpu
            batch_size = min(batch_size, num_gpu * 4)
    if batch_size > num_train_samples // 2:
        batch_size = num_train_samples // 2
    if batch_size < 1:
        batch_size = 1

    brightness_range = 0.5
    contrast_range = 0.5
    saturation_range = 0.5
    saturation = False
    hue = False
    if model_type.startswith('DenseNet') or model_type.startswith('ResNet') \
        or model_type.startswith('Xception') or model_type.startswith('DarkNet') \
        or model_type.startswith('ShuffleNet') or model_type.startswith('MobileNet'):
        if model_type.startswith('MobileNet'):
            lr = (batch_size / 500.0) * 0.1
        else:
            lr = (batch_size / 256.0) * 0.1
        shape = [224, 224]
        save_interval_epochs = 5
        num_epochs = 120
        lr_decay_epochs = [30, 60, 90]
        image_mean = [0.485, 0.456, 0.406]
        image_std = [0.229, 0.224, 0.225]
        brightness_range = 0.9
        contrast_range = 0.9
        saturation_range = 0.9
        brightness = True
        contrast = True
    elif model_type.startswith('YOLOv3') or model_type.startswith('PPYOLO'):
        shape = [608, 608]
        save_interval_epochs = 30
        num_epochs = 270
        lr_decay_epochs = [210, 240]
        lr = 0.001 * batch_size / 64
        image_mean = [0.485, 0.456, 0.406]
        image_std = [0.229, 0.224, 0.225]
        brightness = True
        contrast = True
        saturation = True
        hue = True
        num_steps_each_epoch = num_train_samples // batch_size
        min_warmup_step = max(3 * num_steps_each_epoch, 50 * num_classes)
        if num_gpu == 0:
            num_gpu = 1
        warmup_step = min(min_warmup_step, int(400 * num_classes / num_gpu))
    elif model_type.startswith('FasterRCNN') or model_type.startswith(
            'MaskRCNN'):
        shape = [800, 1333]
        save_interval_epochs = 1
        num_epochs = 12
        lr_decay_epochs = [8, 11]
        image_mean = [0.485, 0.456, 0.406]
        image_std = [0.229, 0.224, 0.225]
        brightness = False
        contrast = False
        lr = 0.02 * batch_size / 16
        num_steps_each_epoch = num_train_samples // batch_size
        min_warmup_step = max(num_steps_each_epoch, 50)
        if num_gpu == 0:
            num_gpu = 1
        warmup_step = min(min_warmup_step, int(4000 / num_gpu))
    elif model_type.startswith('DeepLab') or model_type.startswith('UNet') \
        or model_type.startswith('HRNet_W18') or model_type.startswith('FastSCNN'):
        shape = [512, 512]
        save_interval_epochs = 10
        num_epochs = 100
        lr_decay_epochs = [10, 20]
        image_mean = [0.5, 0.5, 0.5]
        image_std = [0.5, 0.5, 0.5]
        brightness = False
        contrast = False
        lr = 0.01 * batch_size / 2

    params['batch_size'] = batch_size
    params['learning_rate'] = lr
    params["image_shape"] = shape
    params['save_interval_epochs'] = save_interval_epochs
    params['num_epochs'] = num_epochs
    params['lr_decay_epochs'] = lr_decay_epochs
    params['resume_checkpoint'] = None
    params["sensitivities_path"] = None
    params["image_mean"] = image_mean
    params["image_std"] = image_std
    params["horizontal_flip_prob"] = 0.5
    params['brightness'] = brightness
    params["brightness_range"] = brightness_range
    params["brightness_prob"] = 0.5
    params['contrast'] = contrast
    params['contrast_range'] = contrast_range
    params['contrast_prob'] = 0.5
    params['saturation'] = saturation
    params['saturation_range'] = saturation_range
    params['saturation_prob'] = 0.5
    params['hue'] = hue
    params['hue_range'] = 18
    params['hue_prob'] = 0.5
    params['horizontal_flip'] = True

    if model_type in ['YOLOv3', 'PPYOLO', 'FasterRCNN', 'MaskRCNN']:
        num_epochs = params['num_epochs']
        lr_decay_epochs = params['lr_decay_epochs']
        if warmup_step >= lr_decay_epochs[0] * num_steps_each_epoch:
            for i in range(len(lr_decay_epochs)):
                lr_decay_epochs[i] += warmup_step // num_steps_each_epoch
            num_epochs += warmup_step // num_steps_each_epoch
        params['num_epochs'] = num_epochs
        params['lr_decay_epochs'] = lr_decay_epochs
        params['warmup_steps'] = warmup_step

    return params


def get_classification_params(params):
    params["pretrain_weights"] = 'IMAGENET'
    params["lr_policy"] = "Piecewise"
    params['vertical_flip_prob'] = 0.5
    params['vertical_flip'] = True
    params['rotate'] = True
    params['rotate_prob'] = 0.5
    params['rotate_range'] = 30
    return params


def get_detection_params(params):
    params['with_fpn'] = True
    params["pretrain_weights"] = 'IMAGENET'
    if params['model'].startswith('YOLOv3') or params['model'].startswith(
            'PPYOLO'):
        if params['model'].startswith('YOLOv3'):
            params['backbone'] = 'DarkNet53'
        elif params['model'].startswith('PPYOLO'):
            params['backbone'] = 'ResNet50_vd_ssld'
        params['use_mixup'] = True
        params['mixup_alpha'] = 1.5
        params['mixup_beta'] = 1.5
        params['expand_prob'] = 0.5
        params['expand_image'] = True
        params['crop_image'] = True
        params['random_shape'] = True
        params['random_shape_sizes'] = [
            320, 352, 384, 416, 448, 480, 512, 544, 576, 608
        ]
    elif params['model'].startswith('FasterRCNN') or params[
            'model'].startswith('MaskRCNN'):
        params['backbone'] = 'ResNet50'

    return params


def get_segmentation_params(params):
    if params['model'].startswith('DeepLab'):
        params['backbone'] = 'Xception65'
        params["pretrain_weights"] = 'IMAGENET'
    elif params['model'].startswith('UNet'):
        params["pretrain_weights"] = 'COCO'
    elif params['model'].startswith('HRNet_W18') or params['model'].startswith(
            'FastSCNN'):
        params["pretrain_weights"] = 'CITYSCAPES'
    params['loss_type'] = [False, False]
    params['lr_policy'] = 'Polynomial'
    params['optimizer'] = 'SGD'
    params['blur'] = False
    params['blur_prob'] = 0.1
    params['rotate'] = False
    params['max_rotation'] = 15
    params['scale_aspect'] = False
    params['min_ratio'] = 0.5
    params['aspect_ratio'] = 0.33
    params['vertical_flip'] = False
    params['vertical_flip_prob'] = 0.5
    return params


def get_params(data, project_type, num_train_samples, num_classes, num_gpu,
               per_gpu_memory, gpu_list):
    if project_type == "classification":
        if 'model_type' in data:
            model_type = data['model_type']
        else:
            model_type = "MobileNetV2"
        params = get_base_params(model_type, per_gpu_memory, num_train_samples,
                                 num_gpu, gpu_list, num_classes)
        return get_classification_params(params)
    if project_type == "detection":
        if 'model_type' in data:
            model_type = data['model_type']
        else:
            model_type = "YOLOv3"
        params = get_base_params(model_type, per_gpu_memory, num_train_samples,
                                 num_gpu, gpu_list, num_classes)
        return get_detection_params(params)
    if project_type == "instance_segmentation":
        if 'model_type' in data:
            model_type = data['model_type']
        else:
            model_type = "MaskRCNN"
        params = get_base_params(model_type, per_gpu_memory, num_train_samples,
                                 num_gpu, gpu_list, num_classes)
        return get_detection_params(params)
    if project_type == 'segmentation' or project_type == "remote_segmentation":
        if 'model_type' in data:
            model_type = data['model_type']
        else:
            model_type = "DeepLabv3+"
        params = get_base_params(model_type, per_gpu_memory, num_train_samples,
                                 num_gpu, gpu_list, num_classes)
        return get_segmentation_params(params)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import Flask, request, render_template, send_from_directory, jsonify, session, send_file
from werkzeug.utils import secure_filename
from flask_cors import CORS
import argparse
from os import path as osp
import os
import time
import json
import sys
import multiprocessing as mp
from . import workspace_pb2 as w
from .utils import CustomEncoder, ShareData, is_pic, get_logger, TaskStatus, get_ip
import numpy as np

app = Flask(__name__)
CORS(app, supports_credentials=True)
SESSION_TYPE = 'filesystem'
app.config.from_object(__name__)
SD = ShareData()


def init(dirname, logger):
    #初始化工作空间
    from .workspace import init_workspace
    from .system import get_system_info
    SD.workspace = w.Workspace(path=dirname)
    init_workspace(SD.workspace, dirname, logger)
    SD.workspace_dir = dirname
    get_system_info(SD.machine_info)


@app.errorhandler(Exception)
def handle_exception(e):
    ret = {"status": -1, 'message': repr(e)}
    return ret


@app.route('/workspace', methods=['GET', 'PUT'])
def workspace():
    """
    methods=='GET':获取工作目录中项目、数据集、任务的属性
        Args：
            struct(str)：结构类型，可以是'dataset', 'project'或'task'，
            id(str)：结构类型对应的id
            attr_list(list):需要获取的属性的列表
        Return：
            attr(dict):key为属性，value为属性的值，
            status
    methods=='PUT':修改工作目录中项目、数据集、任务的属性
        Args:
            struct(str)：结构类型，可以是'dataset', 'project'或'task'，
            id(str)：结构类型对应的id
            attr_dict(dict):key:需要修改的属性，value：需要修改属性的值
        Return:
            status
    """
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        if data:
            from .workspace import get_attr
            ret = get_attr(data, SD.workspace)
            return ret
        return {'status': 1, 'dirname': SD.workspace_dir}
    if request.method == 'PUT':
        from .workspace import set_attr
        ret = set_attr(data, SD.workspace)
        return ret


@app.route('/dataset', methods=['GET', 'POST', 'PUT', 'DELETE'])
def dataset():
    """
    methods=='GET':获取所有数据集或者单个数据集的信息
        Args：
            did(str, optional):数据集id（可选），如果存在就返回数据集id对应数据集的信息
        Ruturn：
            status
            if 'did' in Args：
                id(str):数据集id，
                dataset_status(int):数据集状态(DatasetStatus)枚举变量的值
                message(str)：数据集状态信息
                attr(dict):数据集属性
            else:
                datasets(list):所有数据集属性的列表

    methods=='POST':创建一个新的数据集
        Args:
            name(str):数据集名字
            desc(str):数据集描述
            dataset_type(str):数据集类型，可以是['classification', 'detection', 'segmentation','instance_segmentation','remote_segmentation']
        Return:
            did(str):数据集id
            status

    methods=='PUT':异步，向数据集导入数据，支持分类、检测、语义分割、实例分割、摇杆分割数据集类型
        Args：
            did(str)：数据集id
            path(str)：数据集路径
        Return:
            status

    methods=='DELETE':删除已有的某个数据集
        Args:
            did(str)：数据集id
        Return:
            status
    """
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        if 'did' in data:
            from .dataset.dataset import get_dataset_status
            ret = get_dataset_status(data, SD.workspace)
            return ret
        from .dataset.dataset import list_datasets
        ret = list_datasets(SD.workspace)
        return ret
    if request.method == 'POST':
        from .dataset.dataset import create_dataset
        ret = create_dataset(data, SD.workspace)
        return ret
    if request.method == 'PUT':
        from .dataset.dataset import import_dataset
        ret = import_dataset(data, SD.workspace, SD.monitored_processes,
                             SD.load_demo_proc_dict)
        return ret

    if request.method == 'DELETE':
        from .dataset.dataset import delete_dataset
        ret = delete_dataset(data, SD.workspace)
        return ret


@app.route('/dataset/details', methods=['GET'])
def dataset_details():
    """
    methods=='GET':获取某个数据集的详细信息
        Args：
            did(str)：数据集id
        Return:
            details(dict)：数据集详细信息,
            status
    """
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        from .dataset.dataset import get_dataset_details
        ret = get_dataset_details(data, SD.workspace)
        return ret


@app.route('/dataset/split', methods=['PUT'])
def dataset_split():
    """
    Args：
        did(str)：数据集id
        val_split(float): 验证集比例
        test_split(float): 测试集比例
    Return:
        status
    """
    data = request.get_json()
    if request.method == 'PUT':
        from .dataset.dataset import split_dataset
        ret = split_dataset(data, SD.workspace)
        return ret


@app.route('/dataset/image', methods=['GET'])
def dataset_img_base64():
    """
    Args:
        GET: 获取图片base64数据，参数：'path' 图片绝对路径
    """
    data = request.get_json()
    if request.method == 'GET':
        from .dataset.dataset import img_base64
        ret = img_base64(data)
        return ret


@app.route('/dataset/file', methods=['GET'])
def get_image_file():
    """
    Args:
        GET: 获取文件数据，参数：'path' 文件绝对路径
    """
    data = request.get_json()
    if request.method == 'GET':
        ret = data['path']
        return send_file(ret)


@app.route('/dataset/npy', methods=['GET'])
def get_npyfile():
    """
    Args:
        GET: 获取文件数据，参数：'path' npy文件绝对路径
    """
    data = request.get_json()
    if request.method == 'GET':
        npy = np.load(data['path'], allow_pickle=True).tolist()
        npy['gt_bbox'] = npy['gt_bbox'].tolist()
        return npy


@app.route('/file', methods=['GET'])
def get_file():
    """
    Args：
        path'(str):文件在服务端的路径
	Return:
		#数据为图片
		img_data(str)： base64图片数据
		status
		#数据为xml文件
		ret:数据流
		#数据为log文件
		ret:json数据
    """
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        path = data['path']
        if not os.path.exists(path):
            return {'status': -1}
        if is_pic(path):
            from .dataset.dataset import img_base64
            ret = img_base64(data, SD.workspace)
            return ret
        file_type = path[(path.rfind('.') + 1):]
        if file_type in ['xml', 'npy', 'log']:
            return send_file(path)
        else:
            pass


@app.route('/project', methods=['GET', 'POST', 'DELETE'])
def project():
    """
    methods=='GET':获取指定项目id的信息
        Args:
            'id'(str, optional):项目id，可选，如果存在就返回项目id对应项目的信息
        Return:
            status,
            if 'id' in Args：
                attr(dict):项目属性
            else:
                projects(list):所有项目属性

    methods=='POST':创建一个项目
        Args:
            name(str): 项目名
            desc(str)：项目描述
            project_type(str)：项目类型
        Return:
            pid(str):项目id
            status

    methods=='DELETE':删除一个项目，以及项目相关的task
        Args：
            pid(str)：项目id
        Return:
            status
    """
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        from .project.project import list_projects
        from .project.project import get_project
        if 'id' in data:
            ret = get_project(data, SD.workspace)
            return ret
        ret = list_projects(SD.workspace)
        return ret
    if request.method == 'POST':
        from .project.project import create_project
        ret = create_project(data, SD.workspace)
        return ret
    if request.method == 'DELETE':
        from .project.project import delete_project
        ret = delete_project(data, SD.workspace)
        return ret


@app.route('/project/task', methods=['GET', 'POST', 'DELETE'])
def task():
    """
    methods=='GET':#获取某个任务的信息或者所有任务的信息
        Args：
            tid(str, optional)：任务id，可选，若存在即返回id对应任务的信息
            resume(str, optional):获取是否可以恢复训练的状态，可选，需在存在tid的情况下才生效
            pid(str, optional)：项目id，可选，若存在即返回该项目id下所有任务信息
        Return:
            status
            if 'tid' in Args：
                task_status(int):任务状态(TaskStatus)枚举变量的值
                message(str)：任务状态信息
                type:任务类型包括{'classification', 'detection', 'segmentation', 'instance_segmentation'}
                resumable(bool):仅Args中存在resume时返回，任务训练是否可以恢复
                max_saved_epochs(int):仅Args中存在resume时返回，当前训练模型保存的最大epoch
            else:
                tasks(list):所有任务属性

    methods=='POST':#创建任务(训练或者裁剪)
        Args:
            pid(str)：项目id
            train(dict)：训练参数
            desc(str, optional):任务描述，可选
            parent_id(str, optional):可选，若存在即表示新建的任务为裁剪任务，parent_id的值为裁剪任务对应的训练任务id
        Return：
            tid(str):任务id
            status

    methods=='DELETE':#删除任务
        Args:
            tid（str）:任务id
        Return:
            status
    """
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        if data:
            if 'pid' not in data:
                from .project.task import get_task_status
                ret = get_task_status(data, SD.workspace)
                return ret
        from .project.task import list_tasks
        ret = list_tasks(data, SD.workspace)
        return ret
    if request.method == 'POST':
        from .project.task import create_task
        ret = create_task(data, SD.workspace)
        return ret
    if request.method == 'DELETE':
        from .project.task import delete_task
        ret = delete_task(data, SD.workspace)
        return ret


@app.route('/project/task/params', methods=['GET', 'POST'])
def task_params():
    """
    methods=='GET':#获取任务id对应的参数，或者获取项目默认参数
        Args:
            tid（str, optional）:获取任务对应的参数
            pid(str，optional)：获取项目对应的默认参数
            model_type(str，optional)：pid存在下有效，对应项目下获取指定模型的默认参数
            gpu_list(list，optional):pid存在下有效，默认值为[0]，使用指定的gpu并获取相应的默认参数
        Return:
            train(dict):训练或者裁剪的参数
            status

    methods=='POST':#设置任务参数，将前端用户设置训练参数dict保存在后端的pkl文件中
        Args:
            tid(str):任务id
            train(dict)：训练参数
        Return:
            status
    """
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        if 'tid' in data:
            from .project.task import get_task_params
            ret = get_task_params(data, SD.workspace)
            ret['train'] = CustomEncoder().encode(ret['train'])
            ret['train'] = json.loads(ret['train'])
            return ret
        if 'pid' in data:
            from .project.task import get_default_params
            ret = get_default_params(data, SD.workspace, SD.machine_info)
            return ret
    if request.method == 'POST':
        from .project.task import set_task_params
        ret = set_task_params(data, SD.workspace)
        return ret


@app.route('/project/task/metrics', methods=['GET'])
def task_metrics():
    """
    methods=='GET':#获取日志数据
        Args:
            tid(str):任务id
            type(str):可以获取日志的类型，[train,eval,sensitivities,prune]，包括训练，评估，敏感度与模型裁剪率关系图，裁剪的日志
        Return:
            status
            if type == 'train':
                train_log(dict): 训练日志
            elif type == 'eval':
                eval_metrics(dict): 评估结果
            elif type == 'sensitivities':
                sensitivities_loss_img(dict): 敏感度与模型裁剪率关系图
            elif type == 'prune':
                prune_log(dict):裁剪日志
    """
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        if data['type'] == 'train':
            from .project.task import get_train_metrics
            ret = get_train_metrics(data, SD.workspace)
            return ret
        if data['type'] == 'eval':
            from .project.task import get_eval_metrics
            ret = get_eval_metrics(data, SD.workspace)
            return ret
        if data['type'] == 'eval_all':
            from .project.task import get_eval_all_metrics
            ret = get_eval_all_metrics(data, SD.workspace)
            return ret
        if data['type'] == 'sensitivities':
            from .project.task import get_sensitivities_loss_img
            ret = get_sensitivities_loss_img(data, SD.workspace)
            return ret
        if data['type'] == 'prune':
            from .project.task import get_prune_metrics
            ret = get_prune_metrics(data, SD.workspace)
            return ret


@app.route('/project/task/train', methods=['POST', 'PUT'])
def task_train():
    """
    methods=='POST':#异步，启动训练或者裁剪任务
        Args：
            tid(str):任务id
            eval_metric_loss(int，optional):可选，裁剪任务时可用，裁剪任务所需的评估loss
        Return:
            status

    methods=='PUT':#改变任务训练的状态，即终止训练或者恢复训练
        Args：
            tid(str)：任务id
            act(str)：[stop,resume]暂停或者恢复
            epoch(int)：(resume下可以设置)恢复训练的起始轮数
        Return:
            status
    """
    data = request.get_json()
    if request.method == 'POST':
        from .project.task import start_train_task
        ret = start_train_task(data, SD.workspace, SD.monitored_processes)
        return ret
    if request.method == 'PUT':
        if data['act'] == 'resume':
            from .project.task import resume_train_task
            ret = resume_train_task(data, SD.workspace, SD.monitored_processes)
            return ret
        if data['act'] == 'stop':
            from .project.task import stop_train_task
            ret = stop_train_task(data, SD.workspace)
            return ret


@app.route('/project/task/train/file', methods=['GET'])
def log_file():
    data = request.get_json()
    if request.method == 'GET':
        path = data['path']
        if not os.path.exists(path):
            return {'status': -1}
        logs = open(path, encoding='utf-8').readlines()
        if len(logs) < 50:
            return {'status': 1, 'log': logs}
        else:
            logs = logs[-50:]
            return {'status': 1, 'log': logs}


@app.route('/project/task/prune', methods=['GET', 'POST', 'PUT'])
def task_prune():
    """
    methods=='GET':#获取裁剪任务的状态
        Args：
            tid(str):任务id
        Return:
            prune_status(int): 裁剪任务状态(PruneStatus)枚举变量的值
            status

    methods=='POST':#异步，创建一个裁剪分析，对于启动裁剪任务前需要先启动裁剪分析
        Args：
            tid(str):任务id
        Return:
            status

    methods=='PUT':#改变裁剪分析任务的状态
        Args：
            tid(str):任务id
            act(str):[stop],目前仅支持停止一个裁剪分析任务
        Return
            status
    """
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        from .project.task import get_prune_status
        ret = get_prune_status(data, SD.workspace)
        return ret
    if request.method == 'POST':
        from .project.task import start_prune_analysis
        ret = start_prune_analysis(data, SD.workspace, SD.monitored_processes)
        return ret
    if request.method == 'PUT':
        if data['act'] == 'stop':
            from .project.task import stop_prune_analysis
            ret = stop_prune_analysis(data, SD.workspace)
        return ret


@app.route('/project/task/evaluate', methods=['GET', 'POST'])
def task_evaluate():
    '''
    methods=='GET':#获取模型评估的结果
        Args：
            tid(str):任务id
        Return:
            evaluate_status(int): 任务状态(TaskStatus)枚举变量的值
            message(str)：描述评估任务的信息
            result(dict)：如果评估成功，返回评估结果的dict，否则为None
            status

    methods=='POST':#异步，创建一个评估任务
        Args：
            tid(str):任务id
            epoch(int,optional):需要评估的epoch，如果为None则会评估训练时指标最好的epoch
            topk(int,optional):分类任务topk指标,如果为None默认输入为5
            score_thresh(float):检测任务类别的score threshhold值，如果为None默认输入为0.5
            overlap_thresh(float):实例分割任务IOU threshhold值，如果为None默认输入为0.3
        Return:
            status
    '''
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        from .project.task import get_evaluate_result
        ret = get_evaluate_result(data, SD.workspace)
        if ret['evaluate_status'] == TaskStatus.XEVALUATED and ret[
                'result'] is not None:
            if 'Confusion_Matrix' in ret['result']:
                ret['result']['Confusion_Matrix'] = ret['result'][
                    'Confusion_Matrix'].tolist()
            ret['result'] = CustomEncoder().encode(ret['result'])
            ret['result'] = json.loads(ret['result'])
        ret['evaluate_status'] = ret['evaluate_status'].value
        return ret
    if request.method == 'POST':
        from .project.task import evaluate_model
        ret = evaluate_model(data, SD.workspace, SD.monitored_processes)
        return ret


@app.route('/project/task/evaluate/file', methods=['GET'])
def task_evaluate_file():
    data = request.get_json()
    if request.method == 'GET':
        if 'path' in data:
            ret = data['path']
            return send_file(ret)
        else:
            from .project.task import get_evaluate_result
            from .project.task import import_evaluate_excel
            ret = get_evaluate_result(data, SD.workspace)
            if ret['evaluate_status'] == TaskStatus.XEVALUATED and ret[
                    'result'] is not None:
                result = ret['result']
                excel_ret = dict()
                excel_ret = import_evaluate_excel(data, result, SD.workspace)
                return excel_ret
            else:
                excel_ret = dict()
                excel_ret['path'] = None
                excel_ret['status'] = -1
                excel_ret['message'] = "评估尚未完成或评估失败"
                return excel_ret


@app.route('/project/task/predict', methods=['GET', 'POST', 'PUT'])
def task_predict():
    '''
    methods=='GET':#获取预测状态
        Args：
            tid(str):任务id
        Return:
            predict_status(int): 预测任务状态(PredictStatus)枚举变量的值
            message(str): 预测信息
            status

    methods=='POST':#创建预测任务，目前仅支持单张图片的预测
        Args:
            tid(str):任务id
            image_data(str):base64编码的image数据
            score_thresh(float，optional):可选，检测任务时有效，检测类别的score threashold值默认是0.5
            epoch(int,float，optional):可选，选择需要做预测的ephoch，默认为评估指标最好的那一个epoch
        Return:
            path(str):服务器上保存预测结果图片的路径
            status
    '''
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        from .project.task import get_predict_status
        ret = get_predict_status(data, SD.workspace)
        return ret
    if request.method == 'POST':
        from .project.task import predict_test_pics
        ret = predict_test_pics(data, SD.workspace, SD.monitored_processes)
        if 'img_list' in data:
            del ret['path']
            return ret
        return ret
    if request.method == 'PUT':
        from .project.task import stop_predict_task
        ret = stop_predict_task(data, SD.workspace)
        return ret


@app.route('/project/task/export', methods=['GET', 'POST', 'PUT'])
def task_export():
    '''
    methods=='GET':#获取导出模型的状态
        Args：
            tid(str):任务id
            quant(str,optional)可选，[log,result]，导出量模型导出状态，若值为log则返回量化的日志；若值为result则返回量化的结果
        Return:
            status
            if quant == 'log':
                quant_log(dict):量化日志
            if quant == 'result'
                quant_result(dict):量化结果
            if quant not in Args：
                export_status(int):模型导出状态(PredictStatus)枚举变量的值
                message(str):模型导出提示信息

    methods=='POST':#导出inference模型或者导出lite模型
        Args:
            tid(str):任务id
            type(str):保存模型的类别[infer,lite]，支持inference模型导出和lite的模型导出
            save_dir(str):保存模型的路径
            epoch(str,optional)可选，指定导出的epoch数默认为评估效果最好的epoch
            quant(bool,optional)可选，type为infer有效，是否导出量化后的模型，默认为False
            model_path(str,optional)可选，type为lite时有效，inference模型的地址
        Return:
            status
            if type == 'infer':
                save_dir:模型保存路径
            if type == 'lite':
                message:模型保存信息

    methods=='PUT':#停止导出模型
        Args:
            tid(str):任务id
        Return:
            export_status(int):模型导出状态(PredictStatus)枚举变量的值
            message(str):停止模型导出提示信息
            status
    '''
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        if 'quant' in data:
            if data['quant'] == 'log':
                from .project.task import get_quant_progress
                ret = get_quant_progress(data, SD.workspace)
                return ret
            if data['quant'] == 'result':
                from .project.task import get_quant_result
                ret = get_quant_result(data, SD.workspace)
                return ret
        from .project.task import get_export_status
        ret = get_export_status(data, SD.workspace)
        ret['export_status'] = ret['export_status'].value
        return ret
    if request.method == 'POST':
        if data['type'] == 'infer':
            from .project.task import export_infer_model
            ret = export_infer_model(data, SD.workspace,
                                     SD.monitored_processes)
            return ret
        if data['type'] == 'lite':
            from .project.task import export_lite_model
            ret = export_lite_model(data, SD.workspace)
            return ret
    if request.method == 'PUT':
        from .project.task import stop_export_task
        stop_export_task(data, SD.workspace)
        return ret


@app.route('/project/task/vdl', methods=['GET'])
def task_vdl():
    '''
    methods=='GET':#打开某个任务的可视化分析工具(VisualDL)
        Args:
            tid(str):任务id
        Return:
            url(str):vdl地址
            status
    '''
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        from .project.task import open_vdl
        ret = open_vdl(data, SD.workspace, SD.current_port,
                       SD.monitored_processes, SD.running_boards)
        return ret


@app.route('/system', methods=['GET', 'DELETE'])
def system():
    '''
    methods=='GET':#获取系统GPU、CPU信息
        Args:
            type(str):[machine_info,gpu_memory_size]选择需要获取的系统信息
        Return:
            status
            if type=='machine_info'
                info(dict):服务端信息
            if type=='gpu_memory_size'
                gpu_mem_infos(list):GPU内存信息
    '''
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        if data['type'] == 'machine_info':
            '''if 'path' not in data:
                data['path'] = None
            from .system import get_machine_info
            ret = get_machine_info(data, SD.machine_info)'''
            from .system import get_system_info
            ret = get_system_info(SD.machine_info)
            return ret

        if data['type'] == 'gpu_memory_size':
            #from .system import get_gpu_memory_size
            from .system import get_gpu_memory_info
            ret = get_gpu_memory_info(SD.machine_info)
            return ret
    if request.method == 'DELETE':
        from .system import exit_system
        ret = exit_system(SD.monitored_processes)
        return ret


@app.route('/demo', methods=['GET', 'POST', 'PUT'])
def demo():
    '''
    methods=='GET':#获取demo下载进度
        Args:
            prj_type(int):项目类型ProjectType枚举变量的int值
        Return:
            status
            attr(dict):demo下载信息

    methods=='POST':#下载或创建demo工程
        Args:
            type(str):{download,load}下载或者创建样例
            prj_type(int):项目类型ProjectType枚举变量的int值
        Return:
            status
            if type=='load':
                did:数据集id
                pid:项目id

    methods=='PUT':#停止下载或创建demo工程
        Args:
            prj_type(int):项目类型ProjectType枚举变量的int值
        Return:
            status
    '''
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        from .demo import get_download_demo_progress
        ret = get_download_demo_progress(data, SD.workspace)
        return ret
    if request.method == 'POST':
        if data['type'] == 'download':
            from .demo import download_demo_dataset
            ret = download_demo_dataset(data, SD.workspace,
                                        SD.load_demo_proc_dict)
            return ret
        if data['type'] == 'load':
            from .demo import load_demo_project
            ret = load_demo_project(data, SD.workspace, SD.monitored_processes,
                                    SD.load_demo_proj_data_dict,
                                    SD.load_demo_proc_dict)
            return ret
    if request.method == 'PUT':
        from .demo import stop_import_demo
        ret = stop_import_demo(data, SD.workspace, SD.load_demo_proc_dict,
                               SD.load_demo_proj_data_dict)
        return ret


@app.route('/model', methods=['GET', 'POST', 'DELETE'])
def model():
    '''
    methods=='GET':#获取一个或者所有模型的信息
        Args:
            mid(str,optional)可选，若存在则返回某个模型的信息
            type(str,optional)可选，[pretrained,exported].若存在则返回对应类型下所有的模型信息
        Return:
            status
            if mid in Args:
                dataset_attr(dict):数据集属性
                task_params(dict):模型训练参数
                eval_result(dict):模型评估结果
            if type in Args and type == 'pretrained':
                pretrained_models(list):所有预训练模型信息
            if type in Args and type == 'exported':
                exported_models(list):所有inference模型的信息

    methods=='POST':#创建一个模型
        Args:
            pid(str):项目id
            tid(str):任务id
            name(str):模型名字
            type(str):创建模型的类型,[pretrained,exported],pretrained代表创建预训练模型、exported代表创建inference或者lite模型
            source_path(str):仅type为pretrained时有效，训练好的模型的路径
            path(str):仅type为exported时有效，inference或者lite模型的路径
            exported_type(int):0为inference模型，1为lite模型
            eval_results(dict，optional):可选，仅type为pretrained时有效，模型评估的指标
        Return:
            status
            if type == 'pretrained':
                pmid(str):预训练模型id
            if type == 'exported':
                emid(str):inference模型id

    methods=='DELETE':删除一个模型
        Args:
            type(str):删除模型的类型，[pretrained,exported]，pretrained代表创建预训练模型、exported代表创建inference或者lite模型
            if type='pretrained':
                pmid:预训练模型id
            if type='exported':
                emid:inference或者lite模型id
        Return:
            status
    '''
    data = request.get_json()
    if data is None:
        data = request.args
    if request.method == 'GET':
        if 'type' in data:
            if data['type'] == 'pretrained':
                from .model import list_pretrained_models
                ret = list_pretrained_models(SD.workspace)
                return ret
            if data['type'] == 'exported':
                from .model import list_exported_models
                ret = list_exported_models(SD.workspace)
                return ret
        from .model import get_model_details
        ret = get_model_details(data, SD.workspace)
        ret['eval_result']['Confusion_Matrix'] = ret['eval_result'][
            'Confusion_Matrix'].tolist()
        ret['eval_result'] = CustomEncoder().encode(ret['eval_result'])
        ret['task_params'] = CustomEncoder().encode(ret['task_params'])
        return ret
    if request.method == 'POST':
        if data['type'] == 'pretrained':
            if 'eval_results' in data:
                data['eval_results']['Confusion_Matrix'] = np.array(data[
                    'eval_results']['Confusion_Matrix'])
            from .model import create_pretrained_model
            ret = create_pretrained_model(data, SD.workspace,
                                          SD.monitored_processes)
            return ret
        if data['type'] == 'exported':
            from .model import create_exported_model
            ret = create_exported_model(data, SD.workspace)
            return ret
    if request.method == 'DELETE':
        if data['type'] == 'pretrained':
            from .model import delete_pretrained_model
            ret = delete_pretrained_model(data, SD.workspace)
            return ret
        if data['type'] == 'exported':
            from .model import delete_exported_model
            ret = delete_exported_model(data, SD.workspace)
            return ret


@app.route('/model/file', methods=['GET'])
def model_file():
    data = request.get_json()
    if request.method == 'GET':
        ret = data['path']
        return send_file(ret)


@app.route('/', methods=['GET'])
def gui():
    if request.method == 'GET':
        file_path = osp.join(
            osp.dirname(__file__), 'templates', 'paddlex_restful_demo.html')
        ip = get_ip()
        url = 'var str_srv_url = "http://' + ip + ':' + str(SD.port) + '";'
        f = open(file_path, 'r+')
        lines = f.readlines()
        for i, line in enumerate(lines):
            if '0.0.0.0:8080' in line:
                lines[i] = url
                break
        f.close()
        f = open(file_path, 'w+')
        f.writelines(lines)
        f.close()
        return render_template('/paddlex_restful_demo.html')


def run(port, workspace_dir):
    if workspace_dir is None:
        user_home = os.path.expanduser('~')
        dirname = osp.join(user_home, "paddlex_workspace")
    else:
        dirname = workspace_dir
    if not osp.exists(dirname):
        os.makedirs(dirname)
    else:
        if not osp.isdir(dirname):
            os.remove(dirname)
            os.makedirs(dirname)
    logger = get_logger(osp.join(dirname, "mcessages.log"))
    init(dirname, logger)
    SD.port = port
    ip = get_ip()
    url = ip + ':' + str(port)
    try:
        logger.info("RESTful服务启动成功后，您可以在浏览器打开 {} 使用WEB版本GUI".format(url))
        app.run(host='0.0.0.0', port=port, threaded=True)
    except:
        print("服务启动不成功，请确保端口号：{}未被防火墙限制".format(port))



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# utils for memory management which is allocated on sharedmemory,
#    note that these structures may not be thread-safe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import time
import math
import struct
import sys
import six

if six.PY3:
    import pickle
else:
    import cPickle as pickle

import json
import uuid
import random
import numpy as np
import weakref
import logging
from multiprocessing import Lock
from multiprocessing import RawArray

logger = logging.getLogger(__name__)


class SharedMemoryError(ValueError):
    """ SharedMemoryError
    """
    pass


class SharedBufferError(SharedMemoryError):
    """ SharedBufferError
    """
    pass


class MemoryFullError(SharedMemoryError):
    """ MemoryFullError
    """

    def __init__(self, errmsg=''):
        super(MemoryFullError, self).__init__()
        self.errmsg = errmsg


def memcopy(dst, src, offset=0, length=None):
    """ copy data from 'src' to 'dst' in bytes
    """
    length = length if length is not None else len(src)
    assert type(dst) == np.ndarray, 'invalid type for "dst" in memcopy'
    if type(src) is not np.ndarray:
        if type(src) is str and six.PY3:
            src = src.encode()
        src = np.frombuffer(src, dtype='uint8', count=len(src))

    dst[:] = src[offset:offset + length]


class SharedBuffer(object):
    """ Buffer allocated from SharedMemoryMgr, and it stores data on shared memory

        note that:
            every instance of this should be freed explicitely by calling 'self.free'
    """

    def __init__(self, owner, capacity, pos, size=0, alloc_status=''):
        """ Init

            Args:
                owner (str): manager to own this buffer
                capacity (int): capacity in bytes for this buffer
                pos (int): page position in shared memory
                size (int): bytes already used
                alloc_status (str): debug info about allocator when allocate this
        """
        self._owner = owner
        self._cap = capacity
        self._pos = pos
        self._size = size
        self._alloc_status = alloc_status
        assert self._pos >= 0 and self._cap > 0, \
            "invalid params[%d:%d] to construct SharedBuffer" \
            % (self._pos, self._cap)

    def owner(self):
        """ get owner
        """
        return SharedMemoryMgr.get_mgr(self._owner)

    def put(self, data, override=False):
        """ put data to this buffer

        Args:
            data (str): data to be stored in this buffer

        Returns:
            None

        Raises:
            SharedMemoryError when not enough space in this buffer
        """
        assert type(data) in [str, bytes], \
            'invalid type[%s] for SharedBuffer::put' % (str(type(data)))
        if self._size > 0 and not override:
            raise SharedBufferError('already has already been setted before')

        if self.capacity() < len(data):
            raise SharedBufferError('data[%d] is larger than size of buffer[%s]'\
                % (len(data), str(self)))

        self.owner().put_data(self, data)
        self._size = len(data)

    def get(self, offset=0, size=None, no_copy=True):
        """ get the data stored this buffer

        Args:
            offset (int): position for the start point to 'get'
            size (int): size to get

        Returns:
            data (np.ndarray('uint8')): user's data in numpy
                which is passed in by 'put'
            None: if no data stored in
        """
        offset = offset if offset >= 0 else self._size + offset
        if self._size <= 0:
            return None

        size = self._size if size is None else size
        assert offset + size <= self._cap, 'invalid offset[%d] '\
            'or size[%d] for capacity[%d]' % (offset, size, self._cap)
        return self.owner().get_data(self, offset, size, no_copy=no_copy)

    def size(self):
        """ bytes of used memory
        """
        return self._size

    def resize(self, size):
        """ resize the used memory to 'size', should not be greater than capacity
        """
        assert size >= 0 and size <= self._cap, \
            "invalid size[%d] for resize" % (size)

        self._size = size

    def capacity(self):
        """ size of allocated memory
        """
        return self._cap

    def __str__(self):
        """ human readable format
        """
        return "SharedBuffer(owner:%s, pos:%d, size:%d, "\
            "capacity:%d, alloc_status:[%s], pid:%d)" \
            % (str(self._owner), self._pos, self._size, \
            self._cap, self._alloc_status, os.getpid())

    def free(self):
        """ free this buffer to it's owner
        """
        if self._owner is not None:
            self.owner().free(self)
            self._owner = None
            self._cap = 0
            self._pos = -1
            self._size = 0
            return True
        else:
            return False


class PageAllocator(object):
    """ allocator used to malloc and free shared memory which
        is split into pages
    """
    s_allocator_header = 12

    def __init__(self, base, total_pages, page_size):
        """ init
        """
        self._magic_num = 1234321000 + random.randint(100, 999)
        self._base = base
        self._total_pages = total_pages
        self._page_size = page_size

        header_pages = int(
            math.ceil((total_pages + self.s_allocator_header) / page_size))

        self._header_pages = header_pages
        self._free_pages = total_pages - header_pages
        self._header_size = self._header_pages * page_size
        self._reset()

    def _dump_alloc_info(self, fname):
        hpages, tpages, pos, used = self.header()

        start = self.s_allocator_header
        end = start + self._page_size * hpages
        try:
            alloc_flags = self._base[start:end].tobytes()
        except:
            alloc_flags = self._base[start:end].tostring()
        info = {
            'magic_num': self._magic_num,
            'header_pages': hpages,
            'total_pages': tpages,
            'pos': pos,
            'used': used
        }
        info['alloc_flags'] = alloc_flags
        fname = fname + '.' + str(uuid.uuid4())[:6]
        with open(fname, 'wb') as f:
            f.write(pickle.dumps(info, -1))
        logger.warn('dump alloc info to file[%s]' % (fname))

    def _reset(self):
        alloc_page_pos = self._header_pages
        used_pages = self._header_pages
        header_info = struct.pack(
            str('III'), self._magic_num, alloc_page_pos, used_pages)
        assert len(header_info) == self.s_allocator_header, \
            'invalid size of header_info'

        memcopy(self._base[0:self.s_allocator_header], header_info)
        self.set_page_status(0, self._header_pages, '1')
        self.set_page_status(self._header_pages, self._free_pages, '0')

    def header(self):
        """ get header info of this allocator
        """
        try:
            header_str = self._base[0:self.s_allocator_header].tobytes()
        except:
            header_str = self._base[0:self.s_allocator_header].tostring()
        magic, pos, used = struct.unpack(str('III'), header_str)

        assert magic == self._magic_num, \
            'invalid header magic[%d] in shared memory' % (magic)
        return self._header_pages, self._total_pages, pos, used

    def empty(self):
        """ are all allocatable pages available
        """
        header_pages, pages, pos, used = self.header()
        return header_pages == used

    def full(self):
        """ are all allocatable pages used
        """
        header_pages, pages, pos, used = self.header()
        return header_pages + used == pages

    def __str__(self):
        header_pages, pages, pos, used = self.header()
        desc = '{page_info[magic:%d,total:%d,used:%d,header:%d,alloc_pos:%d,pagesize:%d]}' \
            % (self._magic_num, pages, used, header_pages, pos, self._page_size)
        return 'PageAllocator:%s' % (desc)

    def set_alloc_info(self, alloc_pos, used_pages):
        """ set allocating position to new value
        """
        memcopy(self._base[4:12],
                struct.pack(str('II'), alloc_pos, used_pages))

    def set_page_status(self, start, page_num, status):
        """ set pages from 'start' to 'end' with new same status 'status'
        """
        assert status in ['0', '1'], 'invalid status[%s] for page status '\
            'in allocator[%s]' % (status, str(self))
        start += self.s_allocator_header
        end = start + page_num
        assert start >= 0 and end <= self._header_size, 'invalid end[%d] of pages '\
            'in allocator[%s]' % (end, str(self))
        memcopy(self._base[start:end], str(status * page_num))

    def get_page_status(self, start, page_num, ret_flag=False):
        start += self.s_allocator_header
        end = start + page_num
        assert start >= 0 and end <= self._header_size, 'invalid end[%d] of pages '\
            'in allocator[%s]' % (end, str(self))
        try:
            status = self._base[start:end].tobytes().decode()
        except:
            status = self._base[start:end].tostring().decode()
        if ret_flag:
            return status

        zero_num = status.count('0')
        if zero_num == 0:
            return (page_num, 1)
        else:
            return (zero_num, 0)

    def malloc_page(self, page_num):
        header_pages, pages, pos, used = self.header()
        end = pos + page_num
        if end > pages:
            pos = self._header_pages
            end = pos + page_num

        start_pos = pos
        flags = ''
        while True:
            # maybe flags already has some '0' pages,
            # so just check 'page_num - len(flags)' pages
            flags = self.get_page_status(pos, page_num, ret_flag=True)

            if flags.count('0') == page_num:
                break

            # not found enough pages, so shift to next few pages
            free_pos = flags.rfind('1') + 1
            pos += free_pos
            end = pos + page_num
            if end > pages:
                pos = self._header_pages
                end = pos + page_num
                flags = ''

            # not found available pages after scan all pages
            if pos <= start_pos and end >= start_pos:
                logger.debug('not found available pages after scan all pages')
                break

        page_status = (flags.count('0'), 0)
        if page_status != (page_num, 0):
            free_pages = self._total_pages - used
            if free_pages == 0:
                err_msg = 'all pages have been used:%s' % (str(self))
            else:
                err_msg = 'not found available pages with page_status[%s] '\
                    'and %d free pages' % (str(page_status), free_pages)
            err_msg = 'failed to malloc %d pages at pos[%d] for reason[%s] and allocator status[%s]' \
                % (page_num, pos, err_msg, str(self))
            raise MemoryFullError(err_msg)

        self.set_page_status(pos, page_num, '1')
        used += page_num
        self.set_alloc_info(end, used)
        return pos

    def free_page(self, start, page_num):
        """ free 'page_num' pages start from 'start'
        """
        page_status = self.get_page_status(start, page_num)
        assert page_status == (page_num, 1), \
            'invalid status[%s] when free [%d, %d]' \
                % (str(page_status), start, page_num)
        self.set_page_status(start, page_num, '0')
        _, _, pos, used = self.header()
        used -= page_num
        self.set_alloc_info(pos, used)


DEFAULT_SHARED_MEMORY_SIZE = 1024 * 1024 * 1024


class SharedMemoryMgr(object):
    """ manage a continouse block of memory, provide
        'malloc' to allocate new buffer, and 'free' to free buffer
    """
    s_memory_mgrs = weakref.WeakValueDictionary()
    s_mgr_num = 0
    s_log_statis = False

    @classmethod
    def get_mgr(cls, id):
        """ get a SharedMemoryMgr with size of 'capacity'
        """
        assert id in cls.s_memory_mgrs, 'invalid id[%s] for memory managers' % (
            id)
        return cls.s_memory_mgrs[id]

    def __init__(self, capacity=None, pagesize=None):
        """ init
        """
        logger.debug('create SharedMemoryMgr')

        pagesize = 64 * 1024 if pagesize is None else pagesize
        assert type(pagesize) is int, "invalid type of pagesize[%s]" \
            % (str(pagesize))

        capacity = DEFAULT_SHARED_MEMORY_SIZE if capacity is None else capacity
        assert type(capacity) is int, "invalid type of capacity[%s]" \
            % (str(capacity))

        assert capacity > 0, '"size of shared memory should be greater than 0'
        self._released = False
        self._cap = capacity
        self._page_size = pagesize

        assert self._cap % self._page_size == 0, \
            "capacity[%d] and pagesize[%d] are not consistent" \
            % (self._cap, self._page_size)
        self._total_pages = self._cap // self._page_size

        self._pid = os.getpid()
        SharedMemoryMgr.s_mgr_num += 1
        self._id = self._pid * 100 + SharedMemoryMgr.s_mgr_num
        SharedMemoryMgr.s_memory_mgrs[self._id] = self
        self._locker = Lock()
        self._setup()

    def _setup(self):
        self._shared_mem = RawArray('c', self._cap)
        self._base = np.frombuffer(
            self._shared_mem, dtype='uint8', count=self._cap)
        self._locker.acquire()
        try:
            self._allocator = PageAllocator(self._base, self._total_pages,
                                            self._page_size)
        finally:
            self._locker.release()

    def malloc(self, size, wait=True):
        """ malloc a new SharedBuffer

        Args:
            size (int): buffer size to be malloc
            wait (bool): whether to wait when no enough memory

        Returns:
            SharedBuffer

        Raises:
            SharedMemoryError when not found available memory
        """
        page_num = int(math.ceil(size / self._page_size))
        size = page_num * self._page_size

        start = None
        ct = 0
        errmsg = ''
        while True:
            self._locker.acquire()
            try:
                start = self._allocator.malloc_page(page_num)
                alloc_status = str(self._allocator)
            except MemoryFullError as e:
                start = None
                errmsg = e.errmsg
                if not wait:
                    raise e
            finally:
                self._locker.release()

            if start is None:
                time.sleep(0.1)
                if ct % 100 == 0:
                    logger.warn('not enough space for reason[%s]' % (errmsg))

                ct += 1
            else:
                break

        return SharedBuffer(self._id, size, start, alloc_status=alloc_status)

    def free(self, shared_buf):
        """ free a SharedBuffer

        Args:
            shared_buf (SharedBuffer): buffer to be freed

        Returns:
            None

        Raises:
            SharedMemoryError when failed to release this buffer
        """
        assert shared_buf._owner == self._id, "invalid shared_buf[%s] "\
            "for it's not allocated from me[%s]" % (str(shared_buf), str(self))
        cap = shared_buf.capacity()
        start_page = shared_buf._pos
        page_num = cap // self._page_size

        #maybe we don't need this lock here
        self._locker.acquire()
        try:
            self._allocator.free_page(start_page, page_num)
        finally:
            self._locker.release()

    def put_data(self, shared_buf, data):
        """  fill 'data' into 'shared_buf'
        """
        assert len(data) <= shared_buf.capacity(), 'too large data[%d] '\
            'for this buffer[%s]' % (len(data), str(shared_buf))
        start = shared_buf._pos * self._page_size
        end = start + len(data)
        assert start >= 0 and end <= self._cap, "invalid start "\
            "position[%d] when put data to buff:%s" % (start, str(shared_buf))
        self._base[start:end] = np.frombuffer(data, 'uint8', len(data))

    def get_data(self, shared_buf, offset, size, no_copy=True):
        """ extract 'data' from 'shared_buf' in range [offset, offset + size)
        """
        start = shared_buf._pos * self._page_size
        start += offset
        if no_copy:
            return self._base[start:start + size]
        else:
            try:
                return self._base[start:start + size].tobytes()
            except:
                return self._base[start:start + size].tostring()

    def __str__(self):
        return 'SharedMemoryMgr:{id:%d, %s}' % (self._id, str(self._allocator))

    def __del__(self):
        if SharedMemoryMgr.s_log_statis:
            logger.info('destroy [%s]' % (self))

        if not self._released and not self._allocator.empty():
            logger.debug('not empty when delete this SharedMemoryMgr[%s]' %
                         (self))
        else:
            self._released = True

        if self._id in SharedMemoryMgr.s_memory_mgrs:
            del SharedMemoryMgr.s_memory_mgrs[self._id]
            SharedMemoryMgr.s_mgr_num -= 1



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path as osp
import random
from .utils import list_files, is_pic, replace_ext, read_seg_ann
import paddlex.utils.logging as logging


def split_seg_dataset(dataset_dir, val_percent, test_percent, save_dir):
    if not osp.exists(osp.join(dataset_dir, "JPEGImages")):
        logging.error("\'JPEGImages\' is not found in {}!".format(dataset_dir))
    if not osp.exists(osp.join(dataset_dir, "Annotations")):
        logging.error("\'Annotations\' is not found in {}!".format(
            dataset_dir))

    all_image_files = list_files(osp.join(dataset_dir, "JPEGImages"))

    image_anno_list = list()
    label_list = list()
    for image_file in all_image_files:
        if not is_pic(image_file):
            continue
        anno_name = replace_ext(image_file, "png")
        if osp.exists(osp.join(dataset_dir, "Annotations", anno_name)):
            image_anno_list.append([image_file, anno_name])
        else:
            anno_name = replace_ext(image_file, "PNG")
            if osp.exists(osp.join(dataset_dir, "Annotations", anno_name)):
                image_anno_list.append([image_file, anno_name])
            else:
                logging.error("The annotation file {} doesn't exist!".format(
                    anno_name))

    if not osp.exists(osp.join(dataset_dir, "labels.txt")):
        for image_anno in image_anno_list:
            labels = read_seg_ann(
                osp.join(dataset_dir, "Annotations", anno_name))
            for i in labels:
                if i not in label_list:
                    label_list.append(i)
        # 如果类标签的最大值大于类别数，添加对应缺失的标签
        if len(label_list) != max(label_list) + 1:
            label_list = [i for i in range(max(label_list) + 1)]

    random.shuffle(image_anno_list)
    image_num = len(image_anno_list)
    val_num = int(image_num * val_percent)
    test_num = int(image_num * test_percent)
    train_num = image_num - val_num - test_num

    train_image_anno_list = image_anno_list[:train_num]
    val_image_anno_list = image_anno_list[train_num:train_num + val_num]
    test_image_anno_list = image_anno_list[train_num + val_num:]

    with open(
            osp.join(save_dir, 'train_list.txt'), mode='w',
            encoding='utf-8') as f:
        for x in train_image_anno_list:
            file = osp.join("JPEGImages", x[0])
            label = osp.join("Annotations", x[1])
            f.write('{} {}\n'.format(file, label))
    with open(
            osp.join(save_dir, 'val_list.txt'), mode='w',
            encoding='utf-8') as f:
        for x in val_image_anno_list:
            file = osp.join("JPEGImages", x[0])
            label = osp.join("Annotations", x[1])
            f.write('{} {}\n'.format(file, label))
    if len(test_image_anno_list):
        with open(
                osp.join(save_dir, 'test_list.txt'), mode='w',
                encoding='utf-8') as f:
            for x in test_image_anno_list:
                file = osp.join("JPEGImages", x[0])
                label = osp.join("Annotations", x[1])
                f.write('{} {}\n'.format(file, label))
    if len(label_list):
        with open(
                osp.join(save_dir, 'labels.txt'), mode='w',
                encoding='utf-8') as f:
            for l in sorted(label_list):
                f.write('{}\n'.format(l))

    return train_num, val_num, test_num
